"""Internal HTTP helper"""
import hashlib
import json
from typing import Optional

import aiohttp

from reduct.error import ReductError


# pylint: disable=too-few-public-methods
class HttpClient:
    """Wrapper for HTTP calls"""

    def __init__(self, url: str, api_token: Optional[str] = None):
        self.url = url
        self.api_token = api_token
        self.headers = {}

    async def request(self, method: str, path: str = "", **kwargs) -> bytes:
        """HTTP request with ReductError exception"""
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method, f"{self.url}{path.strip()}", headers=self.headers, **kwargs
            ) as response:
                if response.ok:
                    return await response.read()

                if response.status == 401:
                    hasher = hashlib.sha256(bytes(self.api_token, "utf-8"))
                    async with session.post(
                        f"{self.url}/auth/refresh",
                        headers={"Authorization": f"Bearer {hasher.hexdigest()}"},
                    ) as auth_resp:
                        if auth_resp.status == 200:
                            data = json.loads(await auth_resp.read())
                            self.headers = {
                                "Authorization": f'Bearer {data["access_token"]}'
                            }
                            return await self.request(method, path, **kwargs)

                raise ReductError(response.status, await response.text())



"""Bucket API"""
import json
from enum import Enum
from typing import Optional, List, Tuple
import time

from pydantic import BaseModel

from reduct.http import HttpClient


class QuotaType(Enum):
    """determines if database has fixed size"""

    NONE = "NONE"
    FIFO = "FIFO"


class BucketSettings(BaseModel):
    """Configuration for a bucket"""

    max_block_size: Optional[int]
    """max block size in bytes"""

    quota_type: Optional[QuotaType]
    """quota type"""

    quota_size: Optional[int]
    """quota size in bytes"""


class BucketInfo(BaseModel):
    """Information about each bucket"""

    name: str
    """name of bucket"""

    entry_count: int
    """number of entries in the bucket"""

    size: int
    """size of bucket data in bytes"""

    oldest_record: int
    """UNIX timestamp of the oldest record in microseconds"""

    latest_record: int
    """UNIX timestamp of the latest record in microseconds"""


class EntryInfo(BaseModel):
    """Entry of bucket"""

    name: str
    """name of entry"""

    size: int
    """size of stored data in bytes"""

    block_count: int
    """number of blocks"""

    record_count: int
    """number of records"""
    oldest_record: int

    """UNIX timestamp of the oldest record in microseconds"""

    latest_record: int
    """UNIX timestamp of the latest record in microseconds"""


class BucketFullInfo(BaseModel):
    """Information about bucket and contained entries"""

    info: BucketInfo
    """statistics about bucket"""

    settings: BucketSettings
    """settings of bucket"""

    entries: List[EntryInfo]
    """information about entries of bucket"""


class Bucket:
    """A bucket of data in Reduct Storage"""

    def __init__(self, name: str, http: HttpClient):
        self._http = http
        self.name = name

    async def get_settings(self) -> BucketSettings:
        """
        Get current settings of bucket
        Returns:
             BucketSettings:
        Raises:
            ReductError: if there is an HTTP error
        """
        return (await self.__get_full_info()).settings

    async def set_settings(self, settings: BucketSettings):
        """
        Update bucket settings
        Args:
            settings: new settings
        Raises:
            ReductError: if there is an HTTP error
        """
        await self._http.request("PUT", f"/b/{self.name}", data=settings.json())

    async def info(self) -> BucketInfo:
        """
        Get statistics about bucket
        Returns:
           BucketInfo:
        Raises:
            ReductError: if there is an HTTP error
        """
        return (await self.__get_full_info()).info

    async def get_entry_list(self) -> List[EntryInfo]:
        """
        Get list of entries with its stats
        Returns:
            List[EntryInfo]
        Raises:
            ReductError: if there is an HTTP error
        """
        return (await self.__get_full_info()).entries

    async def remove(self):
        """
        Remove bucket
        Raises:
            ReductError: if there is an HTTP error
        """
        await self._http.request("DELETE", f"/b/{self.name}")

    async def read(self, entry_name: str, timestamp: Optional[int] = None) -> bytes:
        """
        Read a record from entry
        Args:
            entry_name: name of entry in the bucket
            timestamp: UNIX timestamp in microseconds if None get the latest record
        Returns:
            bytes:
        Raises:
            ReductError: if there is an HTTP error
        """
        params = {"ts": timestamp} if timestamp else None
        return await self._http.request(
            "GET", f"/b/{self.name}/{entry_name}", params=params
        )

    async def write(
        self, entry_name: str, data: bytes, timestamp: Optional[int] = None
    ):
        """
        Write a record to entry
        Args:
            entry_name: name of entry in the bucket
            data: data to write
            timestamp: UNIX timestamp in microseconds. Current time if it's None
        Raises:
            ReductError: if there is an HTTP error

        """
        params = {"ts": timestamp if timestamp else time.time_ns() / 1000}

        await self._http.request(
            "POST",
            f"/b/{self.name}/{entry_name}",
            params=params,
            data=data,
        )

    async def list(
        self, entry_name: str, start: int, stop: int
    ) -> List[Tuple[int, int]]:
        """
        Get list of records in entry for time interval
        Args:
            entry_name: name of entry in the bucket
            start: the beginning of the time interval
            stop: the end of the time interval
        """
        params = {"start": start, "stop": stop}
        data = await self._http.request(
            "GET",
            f"/b/{self.name}/{entry_name}/list",
            params=params,
        )
        records = json.loads(data)["records"]
        items = [(int(record["ts"]), int(record["size"])) for record in records]
        return items

    async def __get_full_info(self) -> BucketFullInfo:
        return BucketFullInfo.parse_raw(
            await self._http.request("GET", f"/b/{self.name}")
        )



"""Reduct Errors"""
from pydantic import BaseModel


class ServerError(BaseModel):
    """sent from the server"""

    detail: str


class ReductError(Exception):
    """General exception for all HTTP errors"""

    def __init__(self, code: int, message: str):
        self._code = code
        self._detail = message
        self.message = ServerError.parse_raw(message) if message else ""
        super().__init__(self.message)

    @property
    def status_code(self):
        """Return HTTP status code"""
        return self._code



"""Main client code"""
from typing import Optional, List

from pydantic import BaseModel

from reduct.bucket import BucketInfo, BucketSettings, Bucket
from reduct.http import HttpClient


class ServerInfo(BaseModel):
    """Server stats"""

    version: str
    """version of the storage in x.y.z format"""

    bucket_count: int
    """number of buckets in the storage"""

    usage: int
    """stored data in bytes"""

    uptime: int
    """storage uptime in seconds"""

    oldest_record: int
    """UNIX timestamp of the oldest record in microseconds"""

    latest_record: int
    """UNIX timestamp of the latest record in microseconds"""


class BucketList(BaseModel):
    """List of buckets"""

    buckets: List[BucketInfo]


class Client:
    """HTTP Client for Reduct Storage HTTP API"""

    def __init__(self, url: str, api_token: Optional[str] = None):
        """
        Constructor

        Args:
            url: URL to connect to the storage
            api_token: API token if the storage uses it for autherization

        Examples:
            >>> client = Client("http://127.0.0.1:8383")
            >>> info = await client.info()
        """
        self._http = HttpClient(url.rstrip("/"), api_token)

    async def info(self) -> ServerInfo:
        """
        Get high level server info

        Returns:
            ServerInfo:

        Raises:
            ReductError: if there is an HTTP error
        """
        return ServerInfo.parse_raw(await self._http.request("GET", "/info"))

    async def list(self) -> List[BucketInfo]:
        """
        Return a list of all buckets on server

        Returns:
            List[BucketInfo]
        Raises:
            ReductError: if there is an HTTP error
        """
        return BucketList.parse_raw(await self._http.request("GET", "/list")).buckets

    async def get_bucket(self, name: str) -> Bucket:
        """
        Load a bucket to work with
        Args:
            name: name of the bucket
        Returns:
            Bucket
        Raises:
            ReductError: if there is an HTTP error
        """
        await self._http.request("HEAD", f"/b/{name}")
        return Bucket(name, self._http)

    async def create_bucket(
        self, name: str, settings: Optional[BucketSettings] = None
    ) -> Bucket:
        """
        Create a new bucket
        Args:
            name: a name for the bucket
            settings: settings for the bucket If None, the server
            default settings is used.
        Returns:
            Bucket: created bucket
        Raises:
            ReductError: if there is an HTTP error
        """
        data = settings.json() if settings else None
        await self._http.request("POST", f"/b/{name}", data=data)
        return Bucket(name, self._http)



# Python implementation of Reduct Storage HTTP API
# (c) 2022 Alexey Timin

""" Setup script
"""
from pathlib import Path

from setuptools import setup, find_packages

PACKAGE_NAME = "reduct-py"
MAJOR_VERSION = 0
MINOR_VERSION = 1
PATCH_VERSION = 0

HERE = Path(__file__).parent.resolve()


def update_package_version(path: Path, version: str):
    """Overwrite/create __init__.py file and fill __version__"""
    with open(path / "VERSION", "w") as version_file:
        version_file.write(f"{version}\n")


def build_version():
    """Build dynamic version and update version in package"""
    version = f"{MAJOR_VERSION}.{MINOR_VERSION}.{PATCH_VERSION}"

    update_package_version(HERE / "pkg" / "reduct", version=version)

    return version


def get_long_description(base_path: Path):
    """Get long package description"""
    return (base_path / "README.md").read_text(encoding="utf-8")


setup(
    name=PACKAGE_NAME,
    version=build_version(),
    description="Reduct Storage Client SDK for Python",
    long_description=get_long_description(HERE),
    long_description_content_type="text/markdown",
    url="https://github.com/reduct-storage/reduct-py",
    author="Alexey Timin, Ciaran Moyne",
    author_email="atimin@gmai.com",
    package_dir={"": "pkg"},
    package_data={"": ["VERSION"]},
    packages=find_packages(where="pkg"),
    python_requires=">=3.8",
    install_requires=["aiohttp~=3.8", "pydantic~=1.9"],
    extras_require={
        "test": ["pytest~=6.2", "pytest-asyncio~=0.18"],
        "lint": [
            "pylint~=2.9",
        ],
        "format": ["black~=22.3"],
        "docs": [
            "mkdocs~=1.3",
            "mkdocs-material~=8.2",
            "plantuml-markdown~=3.5",
            "mkdocs-same-dir~=0.1",
            "mkdocstrings[python]~=0.18",
        ],
    },
)



"""Reduct module"""
from reduct.bucket import QuotaType, Bucket, BucketInfo, BucketSettings, EntryInfo

from reduct.client import (
    Client,
    ServerInfo,
    BucketList,
)

from reduct.error import ReductError
